
"""
Jeffrey Gopez
CHE210D
Final Project
"""

#import python modules
import numpy as np
import time

#import compiled fortran library
import final
#import final

import sys
import os
import atomwrite

#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5
#set the random number seed; useful for debugging
#np.random.seed(342324)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.



def LineSearch(Pos, ChargeList, Kappa, Lambda, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, L, Cut
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [final.calcenergy(Pos + NormDir * x, ChargeList, M, L, Cut, Kappa, Lambda) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = final.calcenergy(Pos + NormDir * dx, ChargeList, M, L, Cut, Kappa, Lambda)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [final.calcenergy(Pos + NormDir * 2. * dx, ChargeList, M, L, Cut, Kappa, Lambda)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [final.calcenergy(Pos + NormDir * Dists[-1], ChargeList, M, L, Cut, Kappa, Lambda)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = final.calcenergy(Pos + NormDir * d3, ChargeList, M, L, Cut, Kappa, Lambda)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = final.calcenergy(Pos + NormDir * d3, ChargeList, M, L, Cut, Kappa, Lambda)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, L, Cut, ChargeList, Kappa, Lambda
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = final.calcenergyforces(Pos, ChargeList, M, L, Cut, Kappa, Lambda, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, ChargeList, Kappa, Lambda, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = final.calcenergyforces(Pos, ChargeList, M, L, Cut, Kappa, Lambda, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    global dt
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    Tau = dt/0.01
    Lambda = np.sqrt(1+(dt/Tau)*(T/InstTemp(Vel) - 1))
    Vel = Vel * Lambda
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, L, Cut, ChargeList, Kappa, Lambda
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = final.calcenergyforces(Pos, ChargeList, M, L, Cut, Kappa, Lambda, Forces)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))


def Charges(N, ABlock, BBlock):
    """Returns charges of each atom
Input: 
    N: number of atoms
    ABlock: Number of Atoms in A block
    BBlock: Number of atoms in B block
Outout:
    ChargeList: list of atom charges
"""
    CperChain = BBlock    
    AtomsPerChain = ABlock + BBlock + CperChain
    NumberChains = N/AtomsPerChain
    
    i=0
    ChargeList = []
    AtomType = []
    while i<NumberChains:
        i += 1        
        
        a = 0
        while a < ABlock:
            ChargeList = np.append(ChargeList, 0)
            AtomType.append("C")
            a += 1
            
        b = 0        
        while b < BBlock:
            ChargeList = np.append(ChargeList, 1)
            AtomType.append("N")
            b += 1
        c = 0
        while c < CperChain:
            ChargeList = np.append(ChargeList, -1)
            AtomType.append("O")
            c+=1
            
    return ChargeList, AtomType

def InitializeSimulation():
    """Returns initial Pos, Vel, PEnergy, and Accel
    """
    global N, L, T, dx, EFracTolLS, EFracTolCG
    Pos = InitPositions(N, L)
    Vel = InitVelocities(N, T)
    PEnergy, Pos = ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG)
    Accel = InitAccel(Pos)
    return Pos, Vel, PEnergy, Accel    

def VVIntegrate(Pos, ChargeList, Vel, Accel):
    # shortcut for calling vvintegrate in fortran file
    global M, L, rc, Kappa, Lambda, dt
    Pos, Vel, Accel, KEnergy, PEnergy = final.vvintegrate(Pos, M, ChargeList, 
                                                          Vel, Accel, L, rc, 
                                                          Kappa, Lambda, dt)
    return Pos, Vel, Accel, KEnergy, PEnergy
    
    
''' Global Parameters'''
dt = .001
T = 1.0
rc = Cut
EFracTolLS=.00000001
EFracTolCG=.0000000001
dx=.001
density = .008
N = 240
L = pow(N/density,1.0/3.)
    
Lambda = 100.
#Electrostatic Strength
Kappa = -1./(L/3.)*np.log(1.e-4*L/(3.*Lambda))
#Screening Length has been set so that the electrostatic interaction 
#between two monomers is 1e-4 at one third of the box length

TotalSteps = 20000000
MovieFrames = 3000
MovieFreq = TotalSteps / MovieFrames
RescaleFreq = 100
SampleFreq = 1000



###### Movie/Figure 1 Generator #######
"""
String = ["M4.pdb","M8.pdb", "M12.pdb", "M16.pdb"]
count = 0
while count < 4:
    print "---------- ",String[count]," ----------"

    ABlock = 3*(count+1)
    BBlock = ABlock/3
    CAtoms = BBlock
    # Total C atoms per chain = to BBlock 
    
    AtomsPerChain = ABlock + BBlock + CAtoms
    #Includes A block + B block + C atoms which are not bonded
    M = AtomsPerChain

    ChargeList, Names = Charges(N,ABlock,BBlock)
    Pos, Vel, PEnergy, Accel = InitializeSimulation()
    #Initializes Pos and Vel, performs conjugate gradient, then inits Accel

    Pdb = atomwrite.pdbfile(String[count], L, AtomNames = Names )
    Pdb.write(Pos)
    #PosCentered = final.centerpos(Pos, L)
    #Pdb.write(PosCentered) <---------- for alternative movie
    #this function moves all of the atoms so 1 molecule remains centered
    

    i = 0
    while i < TotalSteps:
        timer = i * dt
        i = i + 1
        Pos, Vel, Accel, KEnergy, PEnergy = VVIntegrate(Pos, ChargeList, Vel, Accel)
        if i % RescaleFreq == 0:        
            Vel = RescaleVelocities(Vel, T)
            #Rescales Vel using Berendsen Thermostat   
        if i % MovieFreq == 0:
            #PosCentered = final.centerpos(Pos, L)
            #this function moves all of the atoms so 1 molecule remains centered
            #Makes the simulation look weird, but its works            
            #Pdb.write(PosCentered)
            Pdb.write(Pos)
        if i % SampleFreq == 0:
            MIPD = final.calcmipd(Pos, L)
            #MIPD = Mean Interparticle Distance
            print "%03.1f %8.3f" % (timer, MIPD)

    Pdb.close()
    count = count + 1
"""


######## Figure 2 Generator #######
"""
count = 0
SampleFreq = 10
while count < 4:
 
    ABlock = 3*(count+1)
    BBlock = ABlock/3
    CAtoms = BBlock
    # Total C atoms per chain = to BBlock 
    
    AtomsPerChain = ABlock + BBlock + CAtoms
    #Includes A block + B block + C atoms which are not bonded
    M = AtomsPerChain

    ChargeList, Names = Charges(N,ABlock,BBlock)
    
    run = 0
    while run < 5:
    # Runs simulation 5 times for statistics
        
        Pos, Vel, PEnergy, Accel = InitializeSimulation()
        #Initializes Pos and Vel, performs conjugate gradient, then inits Accel        
        MIPD = final.calcmipd(Pos, L)
        
        i = 0.0
        while i < TotalSteps:
            timer = i * dt
            i = i + 1
            Pos, Vel, Accel, KEnergy, PEnergy = VVIntegrate(Pos, ChargeList, Vel, Accel)
            
            if i % RescaleFreq == 0:        
                Vel = RescaleVelocities(Vel, T)
                #Rescales Vel using Berendsen Thermostat                
                
            if i % SampleFreq == 0 and MIPD > 4.25:
                MIPD = final.calcmipd(Pos, L)
            if MIPD < 4.25:
                i = TotalSteps
                # MIPD = Mean Interparticle Distance
        run +=1  
        print "%1.0f %8.4f" % (ABlock+BBlock, timer)
    count = count + 1

""""""

###### Figure 3 and 4 Generator ######
"""
String = ["4_1.pdb","4_2.pdb", "4_3.pdb", "4_4.pdb"]
MIPDlist = [2.9, 3.1, 3.2, 3.5]
# these distance were generated from a initial run, and
# validated in the run used to generate figure 3

SampleFreq = 100
TotalSteps = 10000000000
#This simulation should stop well before this amount of time steps

count = 3
while count < 4:
 
    ABlock = 4
    BBlock = 1+count
    CAtoms = BBlock
    NumChains = 10 
    
    AtomsPerChain = ABlock + BBlock + CAtoms
    #Includes A block + B block + C atoms which are not bonded
    M = AtomsPerChain
    N = M * NumChains
    ChargeList, Names = Charges(N,ABlock,BBlock)
    
############## Figure 3 ##########    
    run = 0
    while run < 5:
    # Runs simulation 5 times for statistics
        
        Pos, Vel, PEnergy, Accel = InitializeSimulation()
        #Initializes Pos and Vel, performs conjugate gradient, then inits Accel        
        MIPD = final.calcmipd(Pos, L)
        EQMIPD = []
        

        #Pdb = atomwrite.pdbfile(String[count], L, AtomNames = Names )
        #Pdb.write(Pos)
        i = 0.0
        while i < 5000.0/dt:
            timer = i * dt
            i = i + 1
            Pos, Vel, Accel, KEnergy, PEnergy = VVIntegrate(Pos, ChargeList, Vel, Accel)            
            if i % RescaleFreq == 0:        
                Vel = RescaleVelocities(Vel, T)
                #Rescales Vel using Berendsen Thermostat                
            
            #if i % MovieFreq == 0:
                #Pdb.write(Pos)            
            
            if i % SampleFreq == 0:
                MIPD = final.calcmipd(Pos, L)

            if MIPD < MIPDlist[count] and i < (TotalSteps - 5000.*SampleFreq):
                i = TotalSteps - 5000.*SampleFreq
                # this makes it so 1500 samples are taken after EQ has been reached
                
            if i % SampleFreq == 0 and i > (TotalSteps - 5000.*SampleFreq):
                EQMIPD = np.append(EQMIPD, MIPD)
                #creates a list of MIPD which will be averaged
                #Pdb.close()
        run +=1  
        print "%1.0f %8.4f %8.4f" % (BBlock, np.mean(EQMIPD), np.std(EQMIPD))
        
############## Figure 4 ##############        
    run = 0
    while run < 5:       
        Pos, Vel, PEnergy, Accel = InitializeSimulation()
        #Initializes Pos and Vel, performs conjugate gradient, then inits Accel        
        MIPD = final.calcmipd(Pos, L)
        
        i = 0.0
        while i < TotalSteps:
            timer = i * dt
            i = i + 1            
            Pos, Vel, Accel, KEnergy, PEnergy = VVIntegrate(Pos, ChargeList, Vel, Accel)
            if i % RescaleFreq == 0:        
                Vel = RescaleVelocities(Vel, T)
                #Rescales Vel using Berendsen Thermostat                
                 
            
            if i % SampleFreq == 0 and MIPD > MIPDlist[count]:
                MIPD = final.calcmipd(Pos, L)
            if MIPD < MIPDlist[count]:
                i = TotalSteps
        run +=1  
        print "%1.0f %8.4f" % (BBlock, timer)
        
    count = count + 1
